@@ -16,12 +16,13 @@
  */
 
 struct hal_ipc_handler {
-	void (*handler) (void *buf, uint16_t len);
+	void (*handler) (void *buf, uint16_t len, int fd);
 	bool var_len;
 	size_t data_len;
 };
 
-bool hal_ipc_init(void);
+bool hal_ipc_init(const char *path, size_t size);
+bool hal_ipc_accept(void);
 void hal_ipc_cleanup(void);
 
 int hal_ipc_cmd(uint8_t service_id, uint8_t opcode, uint16_t len, void *param,
