@@ -134,6 +134,8 @@ extern "C" {
 #define MAP_SVCLASS_ID			0x1134
 #define GNSS_SVCLASS_ID			0x1135
 #define GNSS_SERVER_SVCLASS_ID		0x1136
+#define MPS_SC_SVCLASS_ID		0x113A
+#define MPS_SVCLASS_ID			0x113B
 #define PNP_INFO_SVCLASS_ID		0x1200
 #define GENERIC_NETWORKING_SVCLASS_ID	0x1201
 #define GENERIC_FILETRANS_SVCLASS_ID	0x1202
@@ -150,8 +152,9 @@ extern "C" {
 #define HDP_SVCLASS_ID			0x1400
 #define HDP_SOURCE_SVCLASS_ID		0x1401
 #define HDP_SINK_SVCLASS_ID		0x1402
-#define APPLE_AGENT_SVCLASS_ID		0x2112
+#define GENERIC_ACCESS_SVCLASS_ID	0x1800
 #define GENERIC_ATTRIB_SVCLASS_ID	0x1801
+#define APPLE_AGENT_SVCLASS_ID		0x2112
 
 /*
  * Standard profile descriptor identifiers; note these
@@ -207,6 +210,7 @@ extern "C" {
 #define PBAP_PCE_PROFILE_ID		PBAP_PCE_SVCLASS_ID
 #define PBAP_PSE_PROFILE_ID		PBAP_PSE_SVCLASS_ID
 #define PBAP_PROFILE_ID			PBAP_SVCLASS_ID
+#define MAP_PROFILE_ID			MAP_SVCLASS_ID
 #define PNP_INFO_PROFILE_ID		PNP_INFO_SVCLASS_ID
 #define GENERIC_NETWORKING_PROFILE_ID	GENERIC_NETWORKING_SVCLASS_ID
 #define GENERIC_FILETRANS_PROFILE_ID	GENERIC_FILETRANS_SVCLASS_ID
@@ -223,9 +227,10 @@ extern "C" {
 #define HDP_PROFILE_ID			HDP_SVCLASS_ID
 #define HDP_SOURCE_PROFILE_ID		HDP_SOURCE_SVCLASS_ID
 #define HDP_SINK_PROFILE_ID		HDP_SINK_SVCLASS_ID
-#define APPLE_AGENT_PROFILE_ID		APPLE_AGENT_SVCLASS_ID
-#define GENERIC_ACCESS_PROFILE_ID	0x1800
+#define GENERIC_ACCESS_PROFILE_ID	GENERIC_ACCESS_SVCLASS_ID
 #define GENERIC_ATTRIB_PROFILE_ID	GENERIC_ATTRIB_SVCLASS_ID
+#define APPLE_AGENT_PROFILE_ID		APPLE_AGENT_SVCLASS_ID
+#define MPS_PROFILE_ID			MPS_SC_SVCLASS_ID
 
 /*
  * Compatibility macros for the old MDP acronym
@@ -268,6 +273,10 @@ extern "C" {
 #define SDP_ATTR_GOEP_L2CAP_PSM			0x0200
 #define SDP_ATTR_SVCDB_STATE			0x0201
 
+#define SDP_ATTR_MPSD_SCENARIOS			0x0200
+#define SDP_ATTR_MPMD_SCENARIOS			0x0201
+#define SDP_ATTR_MPS_DEPENDENCIES		0x0202
+
 #define SDP_ATTR_SERVICE_VERSION		0x0300
 #define SDP_ATTR_EXTERNAL_NETWORK		0x0301
 #define SDP_ATTR_SUPPORTED_DATA_STORES_LIST	0x0301
@@ -296,6 +305,8 @@ extern "C" {
 #define SDP_ATTR_SUPPORTED_REPOSITORIES		0x0314
 #define SDP_ATTR_MAS_INSTANCE_ID		0x0315
 #define SDP_ATTR_SUPPORTED_MESSAGE_TYPES	0x0316
+#define SDP_ATTR_PBAP_SUPPORTED_FEATURES	0x0317
+#define SDP_ATTR_MAP_SUPPORTED_FEATURES		0x0317
 
 #define SDP_ATTR_SPECIFICATION_ID		0x0200
 #define SDP_ATTR_VENDOR_ID			0x0201
